<?php

namespace Drupal\gpa_calculator\Form;

use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\AnnounceCommand;
use Drupal\Core\Ajax\HtmlCommand;
use Drupal\Core\Ajax\ReplaceCommand;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\gpa_calculator\Ajax\GpaCalculationCommand;

/**
 * Implements an example form.
 */
class GpaCalculatorForm extends FormBase {

  use StringTranslationTrait;

  /**
   * The default grade settings.
   *
   * @var array
   */
  protected array $defaultGradeSettings = [
    '4.0' => 'A',
    '3.67' => 'A-',
    '3.33' => 'B+',
    '3.0' => 'B',
    '2.67' => 'B-',
    '2.33' => 'C+',
    '2.0' => 'C',
    '1.67' => 'C-',
    '1.33' => 'D+',
    '1.0' => 'D',
    '0.67' => 'D-',
    '0.0' => 'F',
  ];

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'gpa_calculator_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $wrapper_id = $this->getFormId() . '-wrapper';
    $form['#prefix'] = '<div id="' . $wrapper_id . '" aria-live="polite">';
    $form['#suffix'] = '</div>';
    $form['#attributes']['class'][] = 'gpa-calculator-form';

    // Create select options for grades.
    $select_options = ['' => '',];
    foreach ($this->getGradeSettings() as $value) {
      $select_options[$value] = $value;
    }
    $form['#tree'] = TRUE;
    $i = 1;
    $num_rows = $form_state->get('num_rows');

    $form['actions']['#type'] = 'actions';

    $form['gpa_row'] = [
      '#type' => 'container',
      '#prefix' => '<div id="gpa-row-wrapper">',
      '#suffix' => '</div>',
      '#attributes' => [
        'aria-label' => $this->t('GPA Data'),
      ],
    ];

    if (empty($num_rows)) {
      $num_rows = 4;
      $form_state->set('num_rows', $num_rows);
    } else {
      $num_rows = (int) $num_rows;
    }

    for ($i = 0; $i < $num_rows; $i++) {
      $form['gpa_row']['gpa'][$i] = [
        '#type' => 'container',
        '#attributes' => ['class' => ['form-items-inline', 'gpa-row']],
      ];
      $form['gpa_row']['gpa'][$i]['course_name'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Course Name #@num', ['@num' => $i + 1]),
      ];

      // Add a select element for Grade.
      $form['gpa_row']['gpa'][$i]['grade'] = [
        '#type' => 'select',
        '#title' => $this->t('Grade #@num', ['@num' => $i + 1]),
        '#options' => $select_options,
        '#default_value' => '',
      ];

      // Add a number element for Credits Earned.
      $form['gpa_row']['gpa'][$i]['credits_earned'] = [
        '#type' => 'number',
        '#title' => $this->t('Credits Earned #@num', ['@num' => $i + 1]),
        '#min' => 0,
      ];

      if ($i == 0) {
        $form['gpa_row']['actions']['add_row'] = [
          '#type' => 'submit',
          '#value' => t('Add more'),
          '#submit' => ['::addOne'],
          '#limit_validation_errors' => [],
          '#ajax' => [
            'callback' => '::addMoreCallback',
            'wrapper' => 'gpa-row-wrapper',
          ],
          '#attributes' => [
            'aria-label' => $this->t('Add another row of course, grade, and credits'),
          ],
        ];
      }
    }

    $form['prev_gpa'] = [
      '#type' => 'number',
      '#step' => '.01',
      '#title' => $this->t('Cumulative GPA'),
      '#size' => 3,
      '#default_value' => 0,
    ];

    $form['prev_hours'] = [
      '#type' => 'number',
      '#title' => $this->t('Cumulative Credits Earned'),
      '#size' => 3,
      '#default_value' => 0,
    ];

    $form_state->setCached(FALSE);

    // Submit button.
    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Calculate'),
      '#ajax' => [
        'callback' => [$this, 'ajaxSubmitCallback'],
        'wrapper' => $wrapper_id,
        'method' => 'html',
        'disable-refocus' => TRUE,
        'effect' => 'fade',
      ],
    ];

    $form['gpa_results_wrapper'] = [
      '#type' => 'markup',
      '#prefix' => '<div id="gpa-results-wrapper">',
      '#suffix' => '</div>',
      '#markup' => '',
    ];

    $form['#attached']['library'][] = 'gpa_calculator/gpa.calculator.form';
    $form['#attached']['library'][] = 'gpa_calculator/gpa.calculator.admin';

    return $form;
  }

  /**
   * Callback for ajax-enabled add more button.
   */
  public function addMoreCallback(array &$form, FormStateInterface $form_state) {
    // Announce the addition of a new row.
    $response = new AjaxResponse();
    $response->addCommand(new AnnounceCommand('New row of course, grade, and credits input fields added.', 'polite'));
    $response->addCommand(new HtmlCommand('#gpa-row-wrapper', $form['gpa_row']));
    return $response;
  }

  /**
   * Submit handler for the add more button.
   *
   * Increments the max counter and causes a rebuild.
   */
  public function addOne(array &$form, FormStateInterface $form_state): void {
    $name_field = $form_state->get('num_rows');
    $add_button = $name_field + 1;
    $form_state->set('num_rows', $add_button);
    $form_state->setRebuild();
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    // Validate the previous GPA.
    if (!is_numeric($form_state->getValue('prev_gpa'))) {
      $form_state->setErrorByName('prev_gpa', $this->t('Cumulative GPA must be a number.'));
    }
    // Validate the previous credits earned.
    if (!is_numeric($form_state->getValue('prev_hours'))) {
      $form_state->setErrorByName('prev_hours', $this->t('Cumulative Credits Earned must be a number.'));
    }
    // Validate the grades and credits earned.
    $num_rows = $form_state->get('num_rows');
    $valid_courses = 0;

    for ($i = 0; $i < $num_rows; $i++) {
      $row_errors = FALSE;
      $grade = $form_state->getValue(['gpa_row', 'gpa', $i, 'grade']);
      $credits = $form_state->getValue(['gpa_row', 'gpa', $i, 'credits_earned']);

      if (empty($grade) && empty($credits)) {
        continue;
      }

      if (!empty($grade)) {
        // If the letter grade cannot be converted to a numeric grade, set an
        // error.
        $numeric_grade = $this->getNumericGrade($grade);
        if (!$numeric_grade) {
          $form_state->setError($form['gpa_row']['gpa'][$i]['grade'], $this->t('Invalid grade for course #@num.', [
            '@num' => $i+1,
          ]));
          $row_errors = TRUE;
        }
        //
        if (empty($credits)) {
          $form_state->setError($form['gpa_row']['gpa'][$i]['credits_earned'], $this->t('Credits are required for course #@num.', [
            '@num' => $i+1,
          ]));
          $row_errors = TRUE;
        }
      }
      // If the credits are set but the grade is not, set an error.
      if (!empty($credits) && empty($grade)) {
        $form_state->setError($form['gpa_row']['gpa'][$i]['grade'], $this->t('Grade is required for course #@num.', [
          '@num' => $i+1,
        ]));
        $row_errors = TRUE;
      }

      // If there are no errors, increment the valid courses count.
      if (!$row_errors) {
        $valid_courses++;
      }
    }
    // If no valid courses were entered, set an error.
    if ($valid_courses < 1) {
      $form_state->setError($form, $this->t('Please enter at least one grade and credit pair.'));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    // Only perform calculations if there are no errors.
    if (!$form_state->hasAnyErrors()) {

      // Calculate the GPA and output the results.
      $num_rows = $form_state->get('num_rows');
      $total_grade_points = 0.0;
      $total_hours = 0.0;

      // Get the previous GPA and previous credits earned.
      $previous_gpa = $form_state->getValue('prev_gpa');
      $previous_hours = $form_state->getValue('prev_hours');

      // Loop through rows.
      for ($i = 0; $i < $num_rows; $i++) {
        $course_grade = $form_state->getValue(['gpa_row', 'gpa', $i, 'grade']);
        $course_hours = $form_state->getValue(['gpa_row', 'gpa', $i, 'credits_earned']);

        // Skip empty rows.
        if (empty($course_grade) || empty($course_hours)) {
          continue;
        }

        // Convert letter grade to numeric value.
        $numeric_grade = $this->getNumericGrade($course_grade);

        // Increment total hours and grade points.
        $total_hours += (float) $course_hours;
        $total_grade_points += $numeric_grade * $course_hours;
      }

      // Avoid division by zero.
      if ($total_hours > 0) {
        // Calculate Current GPA.
        $current_gpa = $total_grade_points / $total_hours;

        // Calculate Cumulative GPA.
        $previous_grade_points = $previous_gpa * $previous_hours;
        $total_hours += $previous_hours;
        $total_grade_points += $previous_grade_points;
        $cumulative_gpa = $total_grade_points / $total_hours;

        // Output the results.
        $form['gpa_results_wrapper']['#markup'] = $this->t('<p>Current GPA: @current_gpa</p><p>Cumulative GPA: @cumulative_gpa</p>', [
          '@current_gpa' => number_format($current_gpa, 2),
          '@cumulative_gpa' => number_format($cumulative_gpa, 2),
        ]);
      }
    }
  }

  /**
   * AJAX callback for the submit button.
   */
  public function ajaxSubmitCallback(array &$form, FormStateInterface $form_state) {
    $response = new AjaxResponse();

    // If there are any errors, add response commands to announce each one.
    if ($form_state->hasAnyErrors()) {
      foreach ($form_state->getErrors() as $error) {
        $response->addCommand(new AnnounceCommand($this->t('Error: @error', [
          '@error' => $error,
        ]), 'assertive'));
      }
      // Ensure that error messages are printed on the form.
      $form['status_messages'] = [
        '#type' => 'status_messages',
        '#weight' => -10,
      ];
    }

    // Add the custom AJAX command.
    $response->addCommand(new ReplaceCommand(NULL, $form));

    return $response;
  }

  /**
   * Get the grade settings.
   *
   * @return array
   *   The grade settings.
   */
  protected function getGradeSettings(): array {
    $config = $this->config('gpa_calculator.gpa');
    $grades = $config->get('grades');
    if (!empty($grades)) {
      $gradeSettings = [];
      foreach (explode("\n", $grades) as $grade) {
        [$key, $value] = explode('|', $grade, 2);
        $gradeSettings[trim($key)] = trim($value);
      }
      return $gradeSettings;
    }

    return $this->defaultGradeSettings;
  }

  /**
   * Get the numeric grade.
   *
   * @param string $letter_grade
   *   The letter grade.
   *
   * @return bool|string
   *   The numeric grade.
   */
  protected function getNumericGrade($letter_grade): bool|string {
    // Check if gradeSettings has a matching letter grade.
    foreach ($this->getGradeSettings() as $grade_points => $grade) {
      if ($grade === $letter_grade) {
        return $grade_points;
      }
    }

    return FALSE;
  }
}
